#include "nsDebug.h"

#include "nsView.h"

#include <Window.h>
#include <Screen.h>

nsView::nsView() : nsWidget()
{
}


//-------------------------------------------------------------------------
//
// Utility method for implementing both Create(nsIWidget ...) and
// Create(nsNativeWidget...)
//-------------------------------------------------------------------------
nsresult nsView::StandardWindowCreate(nsIWidget *aParent,
                                        const nsRect &aRect,
                                        EVENT_CALLBACK aHandleEventFunction,
                                        nsIDeviceContext *aContext,
                                        nsIAppShell *aAppShell,
                                        nsIToolkit *aToolkit,
                                        nsWidgetInitData *aInitData,
                                        nsNativeWidget aNativeParent)
{
	mIsTopWidgetWindow = PR_FALSE;
	
	BaseCreate(aParent, aRect, aHandleEventFunction, aContext,
	           aAppShell, aToolkit, aInitData);

	mListenForResizes = aNativeParent ? PR_TRUE : aInitData->mListenForResizes;
		
	// Switch to the "main gui thread" if necessary... This method must
	// be executed on the "gui thread"...
	//
	nsToolkit* toolkit = (nsToolkit *)mToolkit;
	if (toolkit && !toolkit->IsGuiThread())
	{
		uint32 args[7];
		args[0] = (uint32)aParent;
		args[1] = (uint32)&aRect;
		args[2] = (uint32)aHandleEventFunction;
		args[3] = (uint32)aContext;
		args[4] = (uint32)aAppShell;
		args[5] = (uint32)aToolkit;
		args[6] = (uint32)aInitData;

		if (nsnull != aParent)
		{
			// nsIWidget parent dispatch
			MethodInfo info(this, this, nsSwitchToUIThread::CREATE, 7, args);
			toolkit->CallMethod(&info);
		}
		else
		{
			// Native parent dispatch
			MethodInfo info(this, this, nsSwitchToUIThread::CREATE_NATIVE, 5, args);
			toolkit->CallMethod(&info);
		}
		return NS_OK;
	}

	mParent = aParent;
	// Useful shortcut, wondering if we can use it also in GetParent() instead nsIWidget* type mParent.
	mWindowParent = (nsWidget *)aParent;
	SetBounds(aRect);

	BRect winrect = BRect(aRect.x, aRect.y, aRect.x + aRect.width - 1, aRect.y + aRect.height - 1);

	// Default mode for window, everything switched off.
	switch (mWindowType)
 	{
		//handle them as childviews until I know better. Shame on me.
		case eWindowType_java:
		case eWindowType_plugin:
			NS_NOTYETIMPLEMENTED("Java and plugin windows not yet implemented properly trying childview"); // to be implemented
 			//These fall thru and behave just like child for the time being. They may require special implementation.
		case eWindowType_child:
 		{
 			//NS_NATIVE_GRAPHIC maybe?
 			//Parent may be a BView if we embed.
			BView *parent= (BView *) (aParent ? aParent->GetNativeData(NS_NATIVE_WIDGET) :  aNativeParent);
			//There seems to be three of these on startup,
			//but I believe that these are because of bugs in 
			//other code as they existed before rewriting this
			//function.
//			NS_PRECONDITION(IsChild(), "Childviews are always created by childwindows?.");
			NS_PRECONDITION(parent, "Childviews without parents don't get added to anything.");
			// A childview that is never added to a parent is very strange.
			if (!parent)
				return NS_ERROR_FAILURE;

			mView = new nsViewBeOS(this, winrect, "Child view", 0, B_WILL_DRAW);
#if defined(BeIME)
			mView->SetFlags(mView->Flags() | B_INPUT_METHOD_AWARE);
#endif	
			bool mustUnlock = parent->Parent() && parent->LockLooper();
 			parent->AddChild(mView);
			if (mustUnlock) parent->UnlockLooper();
			DispatchStandardEvent(NS_CREATE);
			return NS_OK;
 		}

		case eWindowType_invisible:
		case eWindowType_sheet:
		default:
		{
			NS_NOTYETIMPLEMENTED("UNKNOWN or not handled windowtype!!!\n");
		}
	}

	return NS_ERROR_FAILURE;
}


NS_METHOD nsView::WidgetToScreen(const nsRect& aOldRect, nsRect& aNewRect)
{
	BPoint	point;
	point.x = aOldRect.x;
	point.y = aOldRect.y;
	if (mView && mView->LockLooper())
	{
		mView->ConvertToScreen(&point);
		mView->UnlockLooper();
	}
	aNewRect.x = nscoord(point.x);
	aNewRect.y = nscoord(point.y);
	aNewRect.width = aOldRect.width;
	aNewRect.height = aOldRect.height;
	return NS_OK;
}

NS_METHOD nsView::ScreenToWidget(const nsRect& aOldRect, nsRect& aNewRect)
{
	BPoint	point;
	point.x = aOldRect.x;
	point.y = aOldRect.y;
	if (mView && mView->LockLooper())
	{
		mView->ConvertFromScreen(&point);
		mView->UnlockLooper();
	}
	aNewRect.x = nscoord(point.x);
	aNewRect.y = nscoord(point.y);
	aNewRect.width = aOldRect.width;
	aNewRect.height = aOldRect.height;
	return NS_OK;
}


//-------------------------------------------------------------------------
//
// Sanity check potential move coordinates
//
//-------------------------------------------------------------------------
NS_METHOD nsView::ConstrainPosition(PRBool aAllowSlop, PRInt32 *aX, PRInt32 *aY)
{
	//It's undefined for child windows.
	return NS_ERROR_FAILURE;
}

//-------------------------------------------------------------------------
//
// Hide or show this component
//
//-------------------------------------------------------------------------
NS_METHOD nsView::Show(PRBool bState)
{
	if (!mEnabled || bState == mIsVisible)
		return NS_OK;
	

	if(!mView || !mView->LockLooper())
		return NS_ERROR_FAILURE;

	if (bState == PR_TRUE)
		mView->Show();
	else
		mView->Hide();
	
	mView->UnlockLooper();
	mIsVisible = bState;	
	return NS_OK;
}


//-------------------------------------------------------------------------
//
// Move this component
//
//-------------------------------------------------------------------------
nsresult nsView::Move(PRInt32 aX, PRInt32 aY)
{
	//Popups are handled as a view so no need to check if we are popup.
//	if ((mBounds.x == aX) && (mBounds.y == aY))
//	{
		// Nothing to do, since it is already positioned correctly.
//		return NS_OK;    
//	}


	//Comment from old nsWindow:
	// We may reset children visibility here, but it needs special care 
	// - see comment 18 in Bug 311651. More sofisticated code needed.
	if(mView && mView->LockLooper())
	{
		mView->MoveTo(aX, aY);
		mView->UnlockLooper();
	}
	return nsWidget::Move(aX, aY);
}


//-------------------------------------------------------------------------
//
// Resize this component
//
//-------------------------------------------------------------------------
NS_METHOD nsView::Resize(PRInt32 aWidth, PRInt32 aHeight, PRBool aRepaint)
{
//	if (aWidth < 0 || aHeight < 0)
//		return NS_OK;

	if(mView && mView->LockLooper())
	{
		mView->ResizeTo(aWidth - 1, aHeight - 1);
		mView->UnlockLooper();
	}
	return nsWidget::Resize(aWidth, aHeight, aRepaint);
}

//-------------------------------------------------------------------------
//
// Resize this component
//
//-------------------------------------------------------------------------
NS_METHOD nsView::Resize(PRInt32 aX,
                           PRInt32 aY,
                           PRInt32 aWidth,
                           PRInt32 aHeight,
                           PRBool   aRepaint)
{
//	if (aWidth < 0 || aHeight < 0)
//		return NS_OK;
	if(mView && mView->LockLooper())
	{
		mView->MoveTo(aX,aY);
		mView->ResizeTo(aWidth - 1, aHeight - 1);
		mView->UnlockLooper();
	}
	return nsWidget::Resize(aX,aY, aWidth, aHeight, aRepaint);
}


//-------------------------------------------------------------------------
//
// Scroll the bits of a window
//
//-------------------------------------------------------------------------
NS_METHOD nsView::Scroll(PRInt32 aDx, PRInt32 aDy, nsRect *aClipRect)
{
	// Switching trigger on
	mIsScrolling = PR_TRUE;
	//Preventing main view invalidation loop-chain  when children are moving
	//by by hiding children nsWidgets.
	//Maybe this method must be used wider, in move and resize chains
	// and implemented in BeginResizingChildren or in Reset*Visibility() methods
	//Children will be unhidden in ::Update() when called by other than gkview::Scroll() method.
	HideKids(PR_TRUE);
	if (mView && mView->LockLooper())
	{

		BRect src;
		BRect b = mView->Bounds();

		if (aClipRect)
		{
			src.left = aClipRect->x;
			src.top = aClipRect->y;
			src.right = aClipRect->XMost() - 1;
			src.bottom = aClipRect->YMost() - 1;
		}
		else
		{
			src = b;
		}
		// Restricting source by on-screen part of BView
		if (mView->Window())
		{
			BRect screenframe = mView->ConvertFromScreen(BScreen(mView->Window()).Frame());
			src = src & screenframe;
			if (mView->Parent())
			{
				BRect parentframe = mView->ConvertFromParent(mView->Parent()->Frame());
				src = src & parentframe;
			}
		}

		BRegion	invalid;
		invalid.Include(src);
		// Next source clipping check, for same level siblings
		if ( BView *v = mView->Parent() )
		{
			for (BView *child = v->ChildAt(0); child; child = child->NextSibling() )
			{
				BRect siblingframe = mView->ConvertFromParent(child->Frame());
				if (child != mView && child->Parent() != mView)
				{
					invalid.Exclude(siblingframe);
					mView->paintregion.Exclude(siblingframe);
				}
			}
			src = invalid.Frame();
		}

		// make sure we only reference visible bits
		// so we don't trigger a BView invalidate

		if (src.left + aDx < 0)
			src.left = -aDx;
		if (src.right + aDx > b.right)
			src.right = b.right - aDx;
		if (src.top + aDy < 0)
			src.top = -aDy;
		if (src.bottom + aDy > b.bottom)
			src.bottom = b.bottom - aDy;
		
		BRect dest = src.OffsetByCopy(aDx, aDy);
		mView->ConstrainClippingRegion(&invalid);
		// Moving visible content 
		if (src.IsValid() && dest.IsValid())
			mView->CopyBits(src, dest);

		invalid.Exclude(dest);	
		// Native paintregion needs shifting too, it is very important action
		// (as app_server doesn't know about Mozilla viewmanager tricks) -
		// it allows proper update after scroll for areas covered by other windows.
		mView->paintregion.OffsetBy(aDx, aDy);
		mView->ConstrainClippingRegion(&invalid);
		// Time to silently move now invisible children

		nsRect bounds;
		nsWidget *childWidget;
		for (nsIWidget* kid = mFirstChild; kid; kid = kid->GetNextSibling()) 
		{
			childWidget = NS_STATIC_CAST(nsWidget*, kid);
			// No need to Lock/UnlockLooper with GetBounds() and Move() methods
			// using cached values and native MoveBy() instead
			childWidget->GetBounds(bounds);
			bounds.x += aDx;
			bounds.y += aDy; 
			BView *child = ((BView *)kid->GetNativeData(NS_NATIVE_WIDGET));
			if (child)
			{
				//There is native child
				childWidget->SetBounds(bounds);
				mView->paintregion.Exclude(child->Frame());
				child->MoveBy(aDx, aDy);
			}
			else
			{
				childWidget->Move(bounds.x, bounds.y);
			}
		}			

		mView->UnlockLooper();

		// Painting calculated region now,
		// letting Update() to paint remaining content of paintregion
		OnPaint(&invalid);
	}
	return NS_OK;
}


NS_METHOD nsView::SetTitle(const nsAString& aTitle)
{
	if(mView)
		mView->SetName(NS_ConvertUTF16toUTF8(aTitle).get());
	return NS_OK;
}
