#include "nsDebug.h"

#include "nsWindow.h"

#include <Window.h>
#include <Screen.h>

// Used in contrain position.  Specifies how much of a window must remain on screen
#define kWindowPositionSlop 20
// BeOS does not provide this information, so we must hard-code it
#define kWindowBorderWidth 5
#define kWindowTitleBarHeight 24


nsWindow::nsWindow() : nsWidget()
{
	mWindow = NULL; 
}


//-------------------------------------------------------------------------
//
// Utility method for implementing both Create(nsIWidget ...) and
// Create(nsNativeWidget...)
//-------------------------------------------------------------------------
nsresult nsWindow::StandardWindowCreate(nsIWidget *aParent,
                                        const nsRect &aRect,
                                        EVENT_CALLBACK aHandleEventFunction,
                                        nsIDeviceContext *aContext,
                                        nsIAppShell *aAppShell,
                                        nsIToolkit *aToolkit,
                                        nsWidgetInitData *aInitData,
                                        nsNativeWidget aNativeParent)
{

	//Do as little as possible for invisible windows, why are these needed?
	if (mWindowType == eWindowType_invisible)
		return NS_ERROR_FAILURE;
		

	NS_ASSERTION(aInitData->mWindowType != eWindowType_popup || 
		!aParent, "Popups should not be hooked into nsIWidget hierarchy");
	
	mIsTopWidgetWindow = PR_TRUE;
	
	BaseCreate(nsnull, aRect, aHandleEventFunction, aContext,
	           aAppShell, aToolkit, aInitData);

	mListenForResizes = aNativeParent ? PR_TRUE : aInitData->mListenForResizes;
		
	// Switch to the "main gui thread" if necessary... This method must
	// be executed on the "gui thread"...
	//
	nsToolkit* toolkit = (nsToolkit *)mToolkit;
	if (toolkit && !toolkit->IsGuiThread())
	{
		uint32 args[7];
		args[0] = (uint32)aParent;
		args[1] = (uint32)&aRect;
		args[2] = (uint32)aHandleEventFunction;
		args[3] = (uint32)aContext;
		args[4] = (uint32)aAppShell;
		args[5] = (uint32)aToolkit;
		args[6] = (uint32)aInitData;

		if (nsnull != aParent)
		{
			// nsIWidget parent dispatch
			MethodInfo info(this, this, nsSwitchToUIThread::CREATE, 7, args);
			toolkit->CallMethod(&info);
		}
		else
		{
			// Native parent dispatch
			MethodInfo info(this, this, nsSwitchToUIThread::CREATE_NATIVE, 5, args);
			toolkit->CallMethod(&info);
		}
		return NS_OK;
	}

	mParent = aParent;
	// Useful shortcut, wondering if we can use it also in GetParent() instead nsIWidget* type mParent.
	mWindowParent = (nsWidget *)aParent;
	SetBounds(aRect);

	BRect winrect = BRect(aRect.x, aRect.y, aRect.x + aRect.width - 1, aRect.y + aRect.height - 1);

	// Default mode for window, everything switched off.
	uint32 flags = B_NOT_RESIZABLE | B_NOT_MINIMIZABLE | B_NOT_ZOOMABLE | B_NOT_CLOSABLE | B_ASYNCHRONOUS_CONTROLS;
	window_look look = B_NO_BORDER_WINDOW_LOOK;

	if(mWindowType != eWindowType_dialog && mWindowType != eWindowType_toplevel)
	{
		NS_NOTYETIMPLEMENTED("This type of window is not implemented yet.");
		return NS_ERROR_FAILURE;
	}
	
	//eBorderStyle_default is to ask the OS to handle it as it sees best.
	//eBorderStyle_all is same as top_level window default.
	if (eBorderStyle_default == mBorderStyle || eBorderStyle_all & mBorderStyle)
	{
		//(Firefox prefs doesn't go this way, so apparently it wants titlebar, zoom, resize and close.)

		//Look and feel for others are set ok at init.
		if (eWindowType_toplevel==mWindowType)
		{
			look = B_TITLED_WINDOW_LOOK;
			flags = B_ASYNCHRONOUS_CONTROLS;
		}
	}
	else
	{
		if (eBorderStyle_border & mBorderStyle)
			look = B_MODAL_WINDOW_LOOK;

		if (eBorderStyle_resizeh & mBorderStyle)
		{
			//Resize demands at least border
			look = B_MODAL_WINDOW_LOOK;
			flags &= !B_NOT_RESIZABLE;
		}

		//We don't have titlebar menus, so treat like title as it demands titlebar.
		if (eBorderStyle_title & mBorderStyle || eBorderStyle_menu & mBorderStyle)
			look = B_TITLED_WINDOW_LOOK;

		if (eBorderStyle_minimize & mBorderStyle)
			flags &= !B_NOT_MINIMIZABLE;

		if (eBorderStyle_maximize & mBorderStyle)
			flags &= !B_NOT_ZOOMABLE;

		if (eBorderStyle_close & mBorderStyle)
			flags &= !B_NOT_CLOSABLE;
	}

	mWindow = new nsWindowBeOS(this, winrect, "", look, mBWindowFeel, flags);
	if (!mWindow)
		return NS_ERROR_OUT_OF_MEMORY;

	mView = new nsViewBeOS(this, mWindow->Bounds(), "Toplevel view", B_FOLLOW_ALL, 0);

	if (!mView)
		return NS_ERROR_OUT_OF_MEMORY;

	mWindow->AddChild(mView);
	// I'm wondering if we can move part of that code to above
	if (eWindowType_dialog == mWindowType && aParent) 
	{
		nsIWidget *topparent = GetTopParent();
		BWindow* subsetparent = (BWindow *) topparent->GetNativeData(NS_NATIVE_WINDOW);
		if (subsetparent)
		{
			mBWindowFeel = B_FLOATING_SUBSET_WINDOW_FEEL;
			mWindow->SetFeel(mBWindowFeel);
			mWindow->AddToSubset(subsetparent);
		}
	} 
	DispatchStandardEvent(NS_CREATE);
	return NS_OK;

}


NS_METHOD nsWindow::WidgetToScreen(const nsRect& aOldRect, nsRect& aNewRect)
{
	BPoint	point;
	point.x = aOldRect.x;
	point.y = aOldRect.y;
	if (mWindow && mWindow->Lock())
	{

		mWindow->ConvertToScreen(&point);
		mWindow->Unlock();
	}
	aNewRect.x = nscoord(point.x);
	aNewRect.y = nscoord(point.y);
	aNewRect.width = aOldRect.width;
	aNewRect.height = aOldRect.height;
	return NS_OK;
}

NS_METHOD nsWindow::ScreenToWidget(const nsRect& aOldRect, nsRect& aNewRect)
{
	BPoint	point;
	point.x = aOldRect.x;
	point.y = aOldRect.y;
	if (mWindow && mWindow->Lock())
	{
		mWindow->ConvertFromScreen(&point);
		mWindow->Unlock();
	}
	aNewRect.x = nscoord(point.x);
	aNewRect.y = nscoord(point.y);
	aNewRect.width = aOldRect.width;
	aNewRect.height = aOldRect.height;
	return NS_OK;
}


//-------------------------------------------------------------------------
//
// Sanity check potential move coordinates
//
//-------------------------------------------------------------------------
NS_METHOD nsWindow::ConstrainPosition(PRBool aAllowSlop, PRInt32 *aX, PRInt32 *aY)
{
	BScreen screen;
	// If no valid screen, just return
	if (! screen.IsValid()) return NS_OK;
	
	BRect screen_rect = screen.Frame();
	BRect win_bounds = mView->Window()->Frame();

//#ifdef DEBUG_CONSTRAIN_POSITION
//		printf("ConstrainPosition: allowSlop=%s, x=%d, y=%d\n\tScreen :", (aAllowSlop?"T":"F"),*aX,*aY);
//		screen_rect.PrintToStream();
//		printf("\tWindow: ");
//		win_bounds.PrintToStream();
//#endif
		
	if (aAllowSlop) 
	{
		if (*aX < kWindowPositionSlop - win_bounds.IntegerWidth() + kWindowBorderWidth)
			*aX = kWindowPositionSlop - win_bounds.IntegerWidth() + kWindowBorderWidth;
		else if (*aX > screen_rect.IntegerWidth() - kWindowPositionSlop - kWindowBorderWidth)
			*aX = screen_rect.IntegerWidth() - kWindowPositionSlop - kWindowBorderWidth;
			
		if (*aY < kWindowPositionSlop - win_bounds.IntegerHeight() + kWindowTitleBarHeight)
			*aY = kWindowPositionSlop - win_bounds.IntegerHeight() + kWindowTitleBarHeight;
		else if (*aY > screen_rect.IntegerHeight() - kWindowPositionSlop - kWindowBorderWidth)
			*aY = screen_rect.IntegerHeight() - kWindowPositionSlop - kWindowBorderWidth;
			
	} 
	else 
	{
		
		if (*aX < kWindowBorderWidth)
			*aX = kWindowBorderWidth;
		else if (*aX > screen_rect.IntegerWidth() - win_bounds.IntegerWidth() - kWindowBorderWidth)
			*aX = screen_rect.IntegerWidth() - win_bounds.IntegerWidth() - kWindowBorderWidth;
			
		if (*aY < kWindowTitleBarHeight)
			*aY = kWindowTitleBarHeight;
		else if (*aY > screen_rect.IntegerHeight() - win_bounds.IntegerHeight() - kWindowBorderWidth)
			*aY = screen_rect.IntegerHeight() - win_bounds.IntegerHeight() - kWindowBorderWidth;
	}
	return NS_OK;
}


//-------------------------------------------------------------------------
//
// Hide or show this component
//
//-------------------------------------------------------------------------
NS_METHOD nsWindow::Show(PRBool bState)
{
	if (!mEnabled || bState == mIsVisible)
		return NS_OK;
	
	if(!mWindow || !mWindow->Lock())
		return NS_ERROR_FAILURE;

	if (bState == PR_TRUE)
		mWindow->Show();
	else
		mWindow->Hide();
	
	mWindow->Unlock();
	mIsVisible = bState;	
	return NS_OK;
}


//-------------------------------------------------------------------------
//
// Move this component
//
//-------------------------------------------------------------------------
nsresult nsWindow::Move(PRInt32 aX, PRInt32 aY)
{
	//Popups are handled as a view so no need to check if we are popup.
//	if ((mBounds.x == aX) && (mBounds.y == aY))
//	{
		// Nothing to do, since it is already positioned correctly.
//		return NS_OK;    
//	}


	//Comment from old nsWindow:
	// We may reset children visibility here, but it needs special care 
	// - see comment 18 in Bug 311651. More sofisticated code needed.

	mWindow->MoveTo(aX, aY);
	return nsWidget::Move(aX, aY);}


//-------------------------------------------------------------------------
//
// Resize this component
//
//-------------------------------------------------------------------------
NS_METHOD nsWindow::Resize(PRInt32 aWidth, PRInt32 aHeight, PRBool aRepaint)
{
//	if (aWidth < 0 || aHeight < 0)
//		return NS_OK;

	mWindow->ResizeTo(aWidth - 1, aHeight - 1);
	return nsWidget::Resize(aWidth, aHeight, aRepaint);
}

//-------------------------------------------------------------------------
//
// Resize this component
//
//-------------------------------------------------------------------------
NS_METHOD nsWindow::Resize(PRInt32 aX,
                           PRInt32 aY,
                           PRInt32 aWidth,
                           PRInt32 aHeight,
                           PRBool   aRepaint)
{
//	if (aWidth < 0 || aHeight < 0)
//		return NS_OK;

	mWindow->MoveTo(aX,aY);
	mWindow->ResizeTo(aWidth - 1, aHeight - 1);
	return nsWidget::Resize(aX,aY, aWidth, aHeight, aRepaint);
}


//-------------------------------------------------------------------------
//
// Scroll the bits of a window
//
//-------------------------------------------------------------------------
NS_METHOD nsWindow::Scroll(PRInt32 aDx, PRInt32 aDy, nsRect *aClipRect)
{
	// Switching trigger on
	mIsScrolling = PR_TRUE;
	//Preventing main view invalidation loop-chain  when children are moving
	//by by hiding children nsWidgets.
	//Maybe this method must be used wider, in move and resize chains
	// and implemented in BeginResizingChildren or in Reset*Visibility() methods
	//Children will be unhidden in ::Update() when called by other than gkview::Scroll() method.
	HideKids(PR_TRUE);
	if (mView && mView->LockLooper())
	{

		BRect src;
		BRect b = mView->Bounds();

		if (aClipRect)
		{
			src.left = aClipRect->x;
			src.top = aClipRect->y;
			src.right = aClipRect->XMost() - 1;
			src.bottom = aClipRect->YMost() - 1;
		}
		else
		{
			src = b;
		}
		// Restricting source by on-screen part of BView
		if (mView->Window())
		{
			BRect screenframe = mView->ConvertFromScreen(BScreen(mView->Window()).Frame());
			src = src & screenframe;
			if (mView->Parent())
			{
				BRect parentframe = mView->ConvertFromParent(mView->Parent()->Frame());
				src = src & parentframe;
			}
		}

		BRegion	invalid;
		invalid.Include(src);
		// Next source clipping check, for same level siblings
		if ( BView *v = mView->Parent() )
		{
			for (BView *child = v->ChildAt(0); child; child = child->NextSibling() )
			{
				BRect siblingframe = mView->ConvertFromParent(child->Frame());
				if (child != mView && child->Parent() != mView)
				{
					invalid.Exclude(siblingframe);
					mView->paintregion.Exclude(siblingframe);
				}
			}
			src = invalid.Frame();
		}

		// make sure we only reference visible bits
		// so we don't trigger a BView invalidate

		if (src.left + aDx < 0)
			src.left = -aDx;
		if (src.right + aDx > b.right)
			src.right = b.right - aDx;
		if (src.top + aDy < 0)
			src.top = -aDy;
		if (src.bottom + aDy > b.bottom)
			src.bottom = b.bottom - aDy;
		
		BRect dest = src.OffsetByCopy(aDx, aDy);
		mView->ConstrainClippingRegion(&invalid);
		// Moving visible content 
		if (src.IsValid() && dest.IsValid())
			mView->CopyBits(src, dest);

		invalid.Exclude(dest);	
		// Native paintregion needs shifting too, it is very important action
		// (as app_server doesn't know about Mozilla viewmanager tricks) -
		// it allows proper update after scroll for areas covered by other windows.
		mView->paintregion.OffsetBy(aDx, aDy);
		mView->ConstrainClippingRegion(&invalid);
		// Time to silently move now invisible children

		nsRect bounds;
		nsWidget *childWidget;
		for (nsIWidget* kid = mFirstChild; kid; kid = kid->GetNextSibling()) 
		{
			childWidget = NS_STATIC_CAST(nsWidget*, kid);
			// No need to Lock/UnlockLooper with GetBounds() and Move() methods
			// using cached values and native MoveBy() instead
			childWidget->GetBounds(bounds);
			bounds.x += aDx;
			bounds.y += aDy; 
			BView *child = ((BView *)kid->GetNativeData(NS_NATIVE_WIDGET));
			if (child)
			{
				//There is native child
				childWidget->SetBounds(bounds);
				mView->paintregion.Exclude(child->Frame());
				child->MoveBy(aDx, aDy);
			}
			else
			{
				childWidget->Move(bounds.x, bounds.y);
			}
		}			

		mView->UnlockLooper();

		// Painting calculated region now,
		// letting Update() to paint remaining content of paintregion
		OnPaint(&invalid);
	}
	return NS_OK;
}


NS_METHOD nsWindow::SetTitle(const nsAString& aTitle)
{
	mWindow->SetTitle(NS_ConvertUTF16toUTF8(aTitle).get());
	return NS_OK;
}

nsPopupWindow::nsPopupWindow() : nsWindow()
{
	mWindow = NULL; 
}

//-------------------------------------------------------------------------
//
// Utility method for implementing both Create(nsIWidget ...) and
// Create(nsNativeWidget...)
//-------------------------------------------------------------------------
nsresult nsPopupWindow::StandardWindowCreate(nsIWidget *aParent,
                                        const nsRect &aRect,
                                        EVENT_CALLBACK aHandleEventFunction,
                                        nsIDeviceContext *aContext,
                                        nsIAppShell *aAppShell,
                                        nsIToolkit *aToolkit,
                                        nsWidgetInitData *aInitData,
                                        nsNativeWidget aNativeParent)
{
	NS_ASSERTION(aInitData->mWindowType != eWindowType_popup || 
		!aParent, "Popups should not be hooked into nsIWidget hierarchy");
	
	mIsTopWidgetWindow = PR_FALSE;
	
	BaseCreate(aParent, aRect, aHandleEventFunction, aContext,
	           aAppShell, aToolkit, aInitData);

	mListenForResizes = aNativeParent ? PR_TRUE : aInitData->mListenForResizes;
		
	// Switch to the "main gui thread" if necessary... This method must
	// be executed on the "gui thread"...
	//
	nsToolkit* toolkit = (nsToolkit *)mToolkit;
	if (toolkit && !toolkit->IsGuiThread())
	{
		uint32 args[7];
		args[0] = (uint32)aParent;
		args[1] = (uint32)&aRect;
		args[2] = (uint32)aHandleEventFunction;
		args[3] = (uint32)aContext;
		args[4] = (uint32)aAppShell;
		args[5] = (uint32)aToolkit;
		args[6] = (uint32)aInitData;

		if (nsnull != aParent)
		{
			// nsIWidget parent dispatch
			MethodInfo info(this, this, nsSwitchToUIThread::CREATE, 7, args);
			toolkit->CallMethod(&info);
		}
		else
		{
			// Native parent dispatch
			MethodInfo info(this, this, nsSwitchToUIThread::CREATE_NATIVE, 5, args);
			toolkit->CallMethod(&info);
		}
		return NS_OK;
	}

	mParent = aParent;
	// Useful shortcut, wondering if we can use it also in GetParent() instead nsIWidget* type mParent.
	mWindowParent = (nsWidget *)aParent;
	SetBounds(aRect);

	BRect winrect = BRect(aRect.x, aRect.y, aRect.x + aRect.width - 1, aRect.y + aRect.height - 1);

	// Default mode for window, everything switched off.
	uint32 flags = B_NOT_RESIZABLE | B_NOT_MINIMIZABLE | B_NOT_ZOOMABLE | B_NOT_CLOSABLE | B_ASYNCHRONOUS_CONTROLS | B_AVOID_FOCUS | B_NO_WORKSPACE_ACTIVATION;
	window_look look = B_NO_BORDER_WINDOW_LOOK;

//			NS_PRECONDITION(IsChild(), "popups should be? created by childwindows?.");


/* Dunno if popups should allow this at all...
	//eBorderStyle_default is to ask the OS to handle it as it sees best.
	//eBorderStyle_all is same as top_level window default.
	if (eBorderStyle_default == mBorderStyle || eBorderStyle_all & mBorderStyle)
		{
		//(Firefox prefs doesn't go this way, so apparently it wants titlebar, zoom, resize and close.)

		//Look and feel for others are set ok at init.
		if (eWindowType_toplevel==mWindowType)
			{
			look = B_TITLED_WINDOW_LOOK;
			flags = B_ASYNCHRONOUS_CONTROLS;
			}
	}
	else
	{
		if (eBorderStyle_border & mBorderStyle)
			look = B_MODAL_WINDOW_LOOK;

		if (eBorderStyle_resizeh & mBorderStyle)
			{
			//Resize demands at least border
			look = B_MODAL_WINDOW_LOOK;
			flags &= !B_NOT_RESIZABLE;
			}

		//We don't have titlebar menus, so treat like title as it demands titlebar.
		if (eBorderStyle_title & mBorderStyle || eBorderStyle_menu & mBorderStyle)
			look = B_TITLED_WINDOW_LOOK;

		if (eBorderStyle_minimize & mBorderStyle)
			flags &= !B_NOT_MINIMIZABLE;

		if (eBorderStyle_maximize & mBorderStyle)
			flags &= !B_NOT_ZOOMABLE;

		if (eBorderStyle_close & mBorderStyle)
			flags &= !B_NOT_CLOSABLE;
	}

	//popups always avoid focus and don't force the user to another workspace.
	if (eWindowType_popup==mWindowType)
		flags |= B_AVOID_FOCUS | B_NO_WORKSPACE_ACTIVATION;
*/
	mWindow = new nsWindowBeOS(this, winrect, "", look, mBWindowFeel, flags);
	if (!mWindow)
		return NS_ERROR_OUT_OF_MEMORY;

	mView = new nsViewBeOS(this, mWindow->Bounds(), "Popup window", B_FOLLOW_ALL, B_WILL_DRAW);

	if (!mView)
		return NS_ERROR_OUT_OF_MEMORY;

	mWindow->AddChild(mView);
	// I'm wondering if we can move part of that code to above
	if (aNativeParent)
	{
		// Due poor BeOS capability to control windows hierarchy/z-order we use this workaround
		// to show eWindowType_popup (e.g. drop-downs) over floating (subset) parent window.
		if (((BView *)aNativeParent)->Window() &&  ((BView *)aNativeParent)->Window()->IsFloating())
		{
			mBWindowFeel = B_FLOATING_ALL_WINDOW_FEEL;
			mWindow->SetFeel(mBWindowFeel);
		}
	}
	
	DispatchStandardEvent(NS_CREATE);
	return NS_OK;
}

NS_METHOD nsPopupWindow::Show(PRBool bState)
{
	if (bState == PR_TRUE)
		mWindow->SetWorkspaces(B_CURRENT_WORKSPACE);
		
	return nsWindow::Show(bState);
}

